﻿// Created by Brock, Boris, 2015-06-29
// Copyright (c) Boris Brock. All rights reserved.

#region Usings

using System;
using System.Threading;
using System.Timers;
using Timer = System.Timers.Timer;

#endregion

namespace DeskControl.Sources.Serial
{
  internal class ArduinoConnector
  {
    #region Properties

    public int CurrentTableHeight { get; private set; }

    public bool IsConnected
    {
      get { return mPort != null && mPort.IsConnected; }
    }

    #endregion

    #region Members

    private readonly byte[] mByteBuffer = new byte[3];
    private int mByteBufferIndex;
    private SerialPortWrapper mPort;
    private Timer mTimer;

    #endregion

    #region Events

    /// <summary>
    ///   Occurs when the asynchronous connection attempt has succeeded.
    /// </summary>
    public event Action ConnectionSucceeded;

    /// <summary>
    ///   Occurs when the asynchronous connection attempt has failed.
    /// </summary>
    public event Action ConnectionFailed;

    /// <summary>
    ///   Occurs when the table height has changed.
    /// </summary>
    public event Action<int> TableHeightChanged;

    #endregion

    #region Connection Handling

    public void ConnectAsync(string portName)
    {
      if (mPort != null) mPort.Disconnect();

      mPort = new SerialPortWrapper();
      mPort.ConnectionFailed += OnPortConnectionFailed;
      mPort.ConnectionSucceeded += OnPortConnectionSucceeded;
      mPort.BytesReceived += OnPortBytesReceived;
      mPort.ConnectAsync(portName);
    }

    public void Disconnect()
    {
      if (mPort != null) mPort.Disconnect();

      if (mTimer != null)
      {
        mTimer.Stop();
        mTimer.Dispose();
      }
    }

    private void OnPortConnectionFailed()
    {
      if (ConnectionFailed != null) ConnectionFailed();
    }

    private void OnPortConnectionSucceeded()
    {
      SendCommand(Constants.MsgId_GetHeightRequest);

      mTimer = new Timer();
      mTimer.Elapsed += OnTimer;
      mTimer.Interval = 250;
      mTimer.Enabled = true;

      if (ConnectionSucceeded != null) ConnectionSucceeded();
    }

    #endregion

    #region Message Sending

    public void MoveTableDown()
    {
      SendCommand(Constants.MsgId_MoveDown);
    }


    public void MoveTableToHeight(int heightInCm)
    {
      if (heightInCm < CurrentTableHeight)
        SendCommand(Constants.MsgId_SetHeightRequest, (byte) (heightInCm + 1));

      else if (heightInCm > CurrentTableHeight)
        SendCommand(Constants.MsgId_SetHeightRequest, (byte) (heightInCm - 1));
    }

    public void MoveTableUp()
    {
      SendCommand(Constants.MsgId_MoveUp);
    }

    public void MoveUpUntilHeightIsReceived()
    {
      if (!mPort.IsConnected)
        return;

      var moveThread = new Thread(MoveUpUntilHeightIsReceivedThread);
      moveThread.Start();
    }

    private void MoveUpUntilHeightIsReceivedThread()
    {
      // Try to get the height
      SendCommand(Constants.MsgId_GetHeightRequest);
      Thread.Sleep(50);
      if (CurrentTableHeight > 0) return;

      // If height has not yet been determined, move the table
      MoveTableUp();
      int waitTime = 2000;
      while (waitTime > 0)
      {
        if (CurrentTableHeight > 0)
          break;
        Thread.Sleep(100);
        waitTime -= 100;
      }
      StopTable();
    }

    private void OnTimer(object source, ElapsedEventArgs e)
    {
      SendCommand(Constants.MsgId_GetHeightRequest);
    }

    private void SendCommand(byte command, byte param = 0)
    {
      var sequence = new byte[3] {command, param, (byte) (command + param)};
      mPort.SendBytesAsync(sequence);
    }

    public void StopTable()
    {
      SendCommand(Constants.MsgId_StopRequest);
    }

    #endregion

    #region Message Receiving

    private void ByteRead(byte b)
    {
      mByteBuffer[mByteBufferIndex] = b;
      mByteBufferIndex = (mByteBufferIndex + 1)%3;

      if (mByteBufferIndex == 0) MessageRead();
    }

    private byte Checksum(byte b1, byte b2)
    {
      return (byte) (b1 + b2);
    }

    private void MessageRead()
    {
      if (mByteBuffer[0] == Constants.MsgId_GetHeightResponse && mByteBuffer[2] == Checksum(mByteBuffer[0], mByteBuffer[1]))
      {
        CurrentTableHeight = mByteBuffer[1];
        TableHeightChanged(CurrentTableHeight);
      }
    }

    private void OnPortBytesReceived(byte[] bytes)
    {
      foreach (var b in bytes)
        ByteRead(b);
    }

    #endregion
  }
}