﻿// Created by Brock, Boris, 2015-06-30
// Copyright (c) Boris Brock. All rights reserved.

#region Usings

using System;
using System.Collections.Generic;
using System.IO.Ports;
using System.Threading;

#endregion

namespace DeskControl.Sources.Serial
{
  internal class SerialPortWrapper
  {
    #region Delegates

    /// <summary>
    ///   A delegate for byte received event handlers.
    /// </summary>
    /// <param name="bytes">The bytes.</param>
    public delegate void BytesReceivedDelegate(byte[] bytes);

    #endregion

    #region Events

    /// <summary>
    ///   Occurs when bytes have been received from the serial port.
    /// </summary>
    public event BytesReceivedDelegate BytesReceived;

    /// <summary>
    ///   Occurs when the asynchronous connection attempt has succeeded.
    /// </summary>
    public event Action ConnectionSucceeded;

    /// <summary>
    ///   Occurs when the asynchronous connection attempt has failed.
    /// </summary>
    public event Action ConnectionFailed;

    #endregion

    #region Properties

    /// <summary>
    ///   Gets a value indicating whether this serial port is connected.
    /// </summary>
    /// <value>
    ///   <c>true</c> if this instance is connected; otherwise, <c>false</c>.
    /// </value>
    public bool IsConnected { get; private set; }

    #endregion

    #region Members

    /// <summary>
    ///   A queue containing bytes to be sent.
    /// </summary>
    private readonly Queue<byte[]> mTxQueue;

    /// <summary>
    ///   A flag that is used to cancel the thread.
    /// </summary>
    private bool mCancelThread;

    /// <summary>
    ///   The actual serial port.
    /// </summary>
    private SerialPort mPort;

    /// <summary>
    ///   The thread that handles reading and writing.
    /// </summary>
    private Thread mSerialThread;

    #endregion

    #region Public Methods

    /// <summary>
    ///   Initializes a new instance of the <see cref="SerialPortWrapper" /> class.
    /// </summary>
    public SerialPortWrapper()
    {
      mTxQueue = new Queue<byte[]>();
    }

    /// <summary>
    ///   Connects to the given serial port asynchronously.
    /// </summary>
    /// <param name="portName">Name of the port.</param>
    /// <param name="baudrate">The baudrate.</param>
    public void ConnectAsync(string portName, int baudrate = 9600)
    {
      if (IsConnected) Disconnect();

      try
      {
        mPort = new SerialPort(portName, baudrate);
        mPort.DtrEnable = true;
        mPort.Open();

        if (mPort.IsOpen)
        {
          mCancelThread = false;
          mSerialThread = new Thread(SerialThread);
          mSerialThread.Start();

          IsConnected = true;
          if (ConnectionSucceeded != null) ConnectionSucceeded();
          return;
        }
      }
      catch
      {
      }

      if (ConnectionFailed != null) ConnectionFailed();
    }

    /// <summary>
    ///   Disconnects from the serial port.
    /// </summary>
    public void Disconnect()
    {
      if (IsConnected)
      {
        if (mSerialThread != null && mSerialThread.IsAlive)
        {
          mCancelThread = true;
          mSerialThread.Join();
          mSerialThread = null;
        }

        if (mPort != null && mPort.IsOpen)
        {
          mPort.Close();
          mPort.Dispose();
          mPort = null;
        }

        IsConnected = false;
      }
    }

    /// <summary>
    ///   Sends the given bytes asynchronously.
    /// </summary>
    /// <param name="bytes">The bytes.</param>
    public void SendBytesAsync(byte[] bytes)
    {
      if (IsConnected)
        mTxQueue.Enqueue(bytes);
    }

    #endregion

    #region Private Methods

    /// <summary>
    ///   Thread that handles serial communication asynchronously.
    /// </summary>
    private void SerialThread()
    {
      while (!mCancelThread)
      {
        // Rx
        var bytesToRead = mPort.BytesToRead;
        if (bytesToRead > 0)
        {
          var bytes = new byte[bytesToRead];
          mPort.Read(bytes, 0, bytesToRead);
          if (BytesReceived != null) BytesReceived(bytes);
        }

        // Tx
        while (mTxQueue.Count > 0)
        {
          var bytes = mTxQueue.Dequeue();
          mPort.Write(bytes, 0, bytes.Length);
        }

        Thread.Sleep(5);
      }
    }

    #endregion
  }
}